/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "OMS_app.h"
#include <math.h>
#include "Rogus.h"
#include "sched.h"

OMS_basic_app::
OMS_basic_app( long app_sig ) : as( app_sig )
// app_sig should be the application's creator
{
}

//--------------------------------------------------------------------------------------------------

DWORD OMS_app::our_start_time = 0;
static UINT default_output_ioref = 0;
static UINT default_input_ioref = 0;
static HANDLE eventHandle;

static long adjustment;
static long prev_time;
static Bool skip_forward;
static Bool skip_backward;
static long last_mtc;
static long last_mtc_begin;
static long doubleframes_since_mtc_begin;
static int fps;
static HMIDIIN timecode_input;
static Bool eat_timecode;

static void got_mtc()
{
//	tw.pr("got mtc\n");
	long now = OMS_app::time();
	if (now - last_mtc < (6000/fps) && last_mtc_begin != NEVER) // ignore big delays
	{
		doubleframes_since_mtc_begin++;
		if (now - last_mtc_begin < doubleframes_since_mtc_begin*(4000/fps))
		{
//			tw.pr("we need to set clock ahead 1 dms\n");
			skip_forward = True;
			skip_backward = False;
		}
		else if (now - last_mtc_begin > doubleframes_since_mtc_begin*(4000/fps))
		{
//			tw.pr("we need to set clock back 1 dms\n");
			skip_forward = False;
			skip_backward = True;
		}
	}
	else
	{
		last_mtc_begin = now;
		doubleframes_since_mtc_begin = 0;
	}
	last_mtc = now;
}

long OMS_app::time()
{
	long now = (timeGetTime()-our_start_time) * 2;
	if (now > prev_time && skip_forward)
	{
		tw.pr("setting clock ahead 1 dms\n");
		adjustment++;
		skip_forward = False;
	}
	else if (now > prev_time && skip_backward)
	{
		tw.pr("setting clock back 1 dms\n");
		adjustment--;
		skip_backward = False;
	}
	prev_time = now;
	return (now + adjustment);
}

err_ind OMS_app::setTimecodeInput(ioref i, Bool eat)
{
	if (i >= inlh->length)
	{
		ew.w("invalid input node %u",i);
		return err;
	}
	if ((*inlh)[i].open == False)
	{
		ew.w("on \"%s\" (node %d) set timecode input when not open",(*inlh)[i].szPname,i);
		return err;
	}
	timecode_input = (*inlh)[i].handle;
	eat_timecode = eat;
	last_mtc_begin = NEVER;
	return no_err;
}

err_ind OMS_app::unsetTimecodeInput()
{
	timecode_input = NULL;
	return no_err;
}

void
OMS_app::
init_timer()
{
	if (timeBeginPeriod(1) != TIMERR_NOERROR) // set timer resolution to 1 millisecond
		ew.w("Cannot set timer resolution to 1 millisecond");
	our_start_time = timeGetTime();
	adjustment = 0;
	prev_time = 0;
	skip_forward = False;
	skip_backward = False;
	last_mtc = 0;
	fps = 25;
}

c_queue<MyQueueElement> OMS_app::cq( 1000, "" ); // changed name from "input queue" to empty string 11/12/95 PP
parser *OMS_app::lp = 0;

static int binner(int n)
{
	return (int)(pow(2,n));
}

typedef UtilArray<MyMidiIn> MyMidiInArray;
typedef UtilPtr<MyMidiInArray> MyMidiInArrayPtr;
typedef UtilArray<MyMidiOut> MyMidiOutArray;
typedef UtilPtr<MyMidiOutArray> MyMidiOutArrayPtr;

MyMidiInArrayPtr OMS_app::get_input_nodes()
{
	UINT n = midiInGetNumDevs();
	MyMidiInArrayPtr a = new MyMidiInArray(n + (midiator_array->length * Midiator::numInputs()));
	UINT i;
	for (i = 0; i < n; i++)
	{
		(*a)[i].open = 0;
		if (midiInGetDevCaps(i, &((*a)[i]), sizeof(MIDIINCAPS)) != MMSYSERR_NOERROR)
		{
			ew.w("Unable to get info for input device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
		(*a)[i].midiator = NULL;
	}
	int j;
	for (i = n, j = 0; j < midiator_array->length; j++)
	{
		(*midiator_array)[j]->in_start = i;
		UINT k;
		for (k = 0; k < Midiator::numInputs(); k++, i++)
		{
			(*a)[i].open = 0;
			(*midiator_array)[j]->getInputName(k, ((*a)[i].szPname));
			(*a)[i].midiator = (*midiator_array)[j];
		}
	}
	return a;
}

MyMidiOutArrayPtr OMS_app::get_output_nodes()
{
	UINT n = midiOutGetNumDevs();
	MyMidiOutArrayPtr a = new MyMidiOutArray(n + (midiator_array->length * Midiator::numOutputs()));
	UINT i;
	for (i = 0; i < n; i++)
	{
		(*a)[i].open = 0;
		if (midiOutGetDevCaps(i, &((*a)[i]), sizeof(MIDIOUTCAPS)) != MMSYSERR_NOERROR)
		{
			ew.w("Unable to get info for output device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
		(*a)[i].midiator = NULL;
	}
	int j;
	for (i = n, j = 0; j < midiator_array->length; j++)
	{
		(*midiator_array)[j]->out_start = i;
		UINT k;
		for (k = 0; k < Midiator::numOutputs(); k++, i++)
		{
			(*a)[i].open = 0;
			(*midiator_array)[j]->getOutputName(k, ((*a)[i].szPname));
			(*a)[i].midiator = (*midiator_array)[j];
		}
	}
	return a;
}

OMS_app::
OMS_app( long app_sig ) : OMS_basic_app( app_sig ), queueStat(10, binner, "Input Queue Usage")
{
	init_timer();

	eventHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	
	midiators = "1234";
	midiator_array = new MidiatorPtrArray(strlen(midiators));
	UINT i;
	for (i = 0; i < strlen(midiators); i++)
	{
		char buf[6];
		sprintf(buf, "COM%c", midiators[i]);
		(*midiator_array)[i] = new Midiator(buf, my_cdecl_read_hook);
	}

	inlh = get_input_nodes();	
	onlh = get_output_nodes();
//	OMS_node_info<MyMidiIn>::print_nodes(inlh, "input");
//	OMS_node_info<MyMidiOut>::print_nodes(onlh, "output");

	OMS_node_info<MyMidiIn>::select_by_name( default_input, inlh, &default_input_ioref);
	OMS_node_info<MyMidiOut>::select_by_name( default_output, onlh, &default_output_ioref);
}
/*
1. It's not necessary to ever close this connection since OMSSignOut will do it anyway.
3. Though a call to ew.e() will cause an exception to be thrown, meaning that OMS_app's dtor will 
not be called since it was not successfully constructed, stack unwinding will cause OMS_basic_app's 
dtor to be called since by the time we reach the body of OMS_app's ctor, OMS_basic_app's ctor has 
successfully executed.  The upshot of all of this is that OMSSignOut() will be called even if an 
error occurs in OMS_app's ctor.  See Stroustrup's "resource acquisition is initialization" scheme 
on p. 308 of _The C++ Programming Language, 2nd ed._
*/

 OMS_app::
~OMS_app()
{
	closeAllInputs();
	closeAllOutputs();
	if (!CloseHandle(eventHandle))
		ew.w("failed to close eventHandle in OMS_app cleanup");
	timeEndPeriod(1);
}


ioref
OMS_app::
open_input( const char* name )
{
	UINT n = default_input_ioref;
	err_ind e = OMS_node_info<MyMidiIn>::select_by_name( name, inlh, &n);
	if (e)
	{
		ew.w( "couldn't select an input node" ); return 0;
	}
	return open_input((ioref)n);
}

ioref
OMS_app::
open_input()
{
	return open_input(default_input_ioref);
}

ioref
OMS_app::
open_input( ioref i )
{
	UINT n = (UINT) i;
	if ((*inlh)[n].open) // just return the number if already open
		return n;
	// otherwise, we need to open it
	if ((*inlh)[n].midiator == NULL)
	{
		MMRESULT mm = midiInOpen(&((*inlh)[n].handle), n, (DWORD)my_read_hook, NULL, CALLBACK_FUNCTION);
		if (mm != MMSYSERR_NOERROR)
		{
			ew.w( "couldn't open connection to \"%s\" (node %d)", (*inlh)[n].szPname, n ); return (ioref)n;
		}
		mm = midiInStart((*inlh)[n].handle);
		if (mm != MMSYSERR_NOERROR)
		{
			ew.w( "couldn't start connection to \"%s\" (node %d)", (*inlh)[n].szPname, n ); return (ioref)n;
		}
		(*inlh)[n].sysex = new MySysexInputBuffer[my_sysex_num_buffers];
		(*inlh)[n].assembly_position = 0;
		int j;
		for (j = 0; j < my_sysex_num_buffers; j++)
		{
			MIDIHDR* hdr = &((*inlh)[n].sysex[j].hdr);
			hdr->lpData = (char*)((*inlh)[n].sysex[j].buf);
			hdr->dwBufferLength = my_sysex_input_size;
			hdr->dwFlags = 0;
			if (midiInPrepareHeader((*inlh)[n].handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
				ew.w("on \"%s\" (node %d) unable to prepare sysex input buffer", (*inlh)[n].szPname, n );
			if (midiInAddBuffer((*inlh)[n].handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
				ew.w("on \"%s\" (node %d) unable to add sysex input buffer", (*inlh)[n].szPname, n );
		}
	}
	else
		(*inlh)[n].midiator->openInput(n-((*inlh)[n].midiator->in_start));
	(*inlh)[n].open = 1; // remember that we opened it
	return (ioref)n;
}

ioref
OMS_app::
open_output( const char* name )
{
	UINT n = default_output_ioref;;
	err_ind e = OMS_node_info<MyMidiOut>::select_by_name( name, onlh, &n);
	if (e)
	{
		ew.w( "couldn't select an output node" ); return 0;
	}
	return open_output((ioref)n);
}

ioref
OMS_app::
open_output()
{
	return open_output(default_output_ioref);
}

ioref
OMS_app::
open_output( ioref i )
{
	UINT n = (UINT) i;
	(*onlh)[n].running_status = 0; // clear out running status
	if ((*onlh)[n].open) // just return the number if already open
		return n;
	// otherwise, we need to open it
	if ((*onlh)[n].midiator == NULL)
	{
		MMRESULT mm = midiOutOpen(&((*onlh)[n].handle), n, (DWORD)my_read_hook, NULL, CALLBACK_FUNCTION);
		if (mm != MMSYSERR_NOERROR)
		{
			ew.w( "couldn't open connection to \"%s\" (node %d)", (*onlh)[n].szPname, n ); return (ioref)n;
		}
	}
	else
		(*onlh)[n].midiator->openOutput(n-((*onlh)[n].midiator->out_start));
	(*onlh)[n].open = 1; // remember that we opened it
	return (ioref)n;
}

void
OMS_app::
closeAllInputs()
{
	for (int i = 0; i < inlh->length; i++)
		if ((*inlh)[i].open)
		{
			if ((*inlh)[i].midiator != NULL)
			{
				(*inlh)[i].midiator->closeInput(i - ((*inlh)[i].midiator->in_start));
				(*inlh)[i].open = 0;
			}
			else
			{
				if (midiInReset((*inlh)[i].handle) != MMSYSERR_NOERROR)
					ew.w("on \"%s\" (node %d) midiInReset failed", (*inlh)[i].szPname, i);
				else if ((*inlh)[i].sysex != NULL)
				{
					int j;
					for (j = 0; j < my_sysex_num_buffers; j++)
					{
						MIDIHDR* hdr = &((*inlh)[i].sysex[j].hdr);
						if (midiInUnprepareHeader((*inlh)[i].handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
							ew.w("on \"%s\" (node %d) unable to unprepare sysex input buffer", (*inlh)[i].szPname, i );
					}
					delete[]((*inlh)[i].sysex);
					(*inlh)[i].sysex = NULL;
				}
				if ((midiInStop((*inlh)[i].handle) == MMSYSERR_NOERROR) && (midiInClose((*inlh)[i].handle) == MMSYSERR_NOERROR))
					(*inlh)[i].open = 0;
				else
					ew.w("unable to close \"%s\" (node %d)", (*inlh)[i].szPname, i);
			}
		}
}

void
OMS_app::
closeAllOutputs()
{
	for (int i = 0; i < onlh->length; i++)
		if ((*onlh)[i].open)
		{
			if ((*onlh)[i].midiator != NULL)
			{
				(*onlh)[i].midiator->closeOutput(i - ((*onlh)[i].midiator->out_start));
				(*onlh)[i].open = 0;
			}
			else if (midiOutClose((*onlh)[i].handle) == MMSYSERR_NOERROR)
				(*onlh)[i].open = 0;
			else
				ew.w("unable to close \"%s\" (node %d)", (*onlh)[i].szPname, i);
		}
}

int
OMS_app::
howManyOutputNodes()
{
	return onlh->length;
}

int
OMS_app::
howManyInputNodes()
{
	return inlh->length;
}

const char*
OMS_app::
getOutputNodeName(ioref i)
{
	if (i >= onlh->length)
	{
		ew.w("invalid output node %u",i);
		return "invalid";
	}
	return (*onlh)[i].szPname;
}

const char*
OMS_app::
getInputNodeName(ioref i)
{
	if (i >= inlh->length)
	{
		ew.w("invalid input node %u",i);
		return "invalid";
	}
	return (*inlh)[i].szPname;
}

/*
ioref
OMS_app::
getOutputRefnum(int i)
{
	return (ioref)i;
}
*/

void
OMS_app::
my_cdecl_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	my_read_hook(hMidiIn, wMsg, dwInstance, dwParam1, dwParam2);
}

void CALLBACK
OMS_app::
my_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	if (wMsg == MIM_DATA)
	{
		if (hMidiIn == timecode_input && timecode_input != NULL)
		{
			if ((dwParam1 & 0xf0ff) == 0xf1)
			{
				got_mtc();
			}
			else if ((dwParam1 & 0xf0ff) == 0x70f1)
			{
				switch ((dwParam1 >> 8) & 0xf)
				{
				case 0:
					fps = 24;
					break;
				case 1:
					fps = 25;
					break;
				default:
					fps = 30;
					break;
				}
			}
			if ((dwParam1 & 0xff) == 0xf1 && eat_timecode == True)
				return;
		}
		if (lp && lp->handles(dwParam1))
		{
			UINT n = 0;
			if (hMidiIn == NULL)
				n = (UINT)dwParam2; // Midiator passes its ioref in dwParam2
			else
			{
				for (int i = 0; i < oap->inlh->length; i++)
					if ((*(oap->inlh))[i].open && (*(oap->inlh))[i].handle == hMidiIn)
						n = (UINT)i;
			}
			MyQueueElement q;
			q.dwParam1 = dwParam1;
			q.dwParam2 = dwParam2;
			q.n = n;
			q.hdr = NULL;
			cq.insertp(&q);
			schedp->processQueue(); // ask schedule thread to call our processQueue
		}
	}
	else if (wMsg == MIM_LONGDATA)
	{
		MIDIHDR* hdr = (MIDIHDR*) dwParam1;
		UINT n = 0;
		for (int i = 0; i < oap->inlh->length; i++)
			if ((*(oap->inlh))[i].open && (*(oap->inlh))[i].handle == hMidiIn)
				n = (UINT)i;
		MyQueueElement q;
		q.hdr = hdr;
		q.n = n;
		cq.insertp(&q);
		schedp->processQueue();
	}
	else if (wMsg == MOM_DONE)
		SetEvent(eventHandle); // wake up write_ed(), waiting on sysex
	else if (wMsg == MIM_ERROR || wMsg == MIM_LONGERROR)
	{
		UINT n = 0;
		for (int i = 0; i < oap->inlh->length; i++)
			if ((*(oap->inlh))[i].open && (*(oap->inlh))[i].handle == hMidiIn)
				n = (UINT)i;
		if (wMsg == MIM_ERROR)
			ew.w("on \"%s\" (node %d) input error", (*(oap->inlh))[n].szPname, n);
		else
			ew.w("on \"%s\" (node %d) sysex input error", (*(oap->inlh))[n].szPname, n);
	}
}

void OMS_app::
processQueue()
{
	MyQueueElement *q;
	int i = 0;
	while ( (q = cq.headp()) != NULL )
	{
		if (q->hdr != NULL)
		{
//			tw.pr("Processing queue for sysex\n");
//			for (unsigned int foo = 0; foo < q->hdr->dwBytesRecorded; foo++)
//				tw.pr("%02x ", (uchar)q->hdr->lpData[foo]);
//			tw.pr("\n");
			(*inlh)[q->n].assembly_order[(*inlh)[q->n].assembly_position++] = q->hdr;
			if ((uchar)q->hdr->lpData[q->hdr->dwBytesRecorded-1] == 0xf7) // if end-of-sysex
			{
//				tw.pr("Got an end-of-sysex\n");
				int len = 0;
				int j;
				for (j = 0; j < (*inlh)[q->n].assembly_position; j++)
					len += (*inlh)[q->n].assembly_order[j]->dwBytesRecorded;
				uchar* tmp = new uchar[len];
				uchar* p = tmp;
				for (j = 0; j < (*inlh)[q->n].assembly_position; j++)
				{
					unsigned int k;
					for (k = 0; k < (*inlh)[q->n].assembly_order[j]->dwBytesRecorded; k++)
						*(p++) = (*inlh)[q->n].assembly_order[j]->lpData[k];
					if (midiInAddBuffer((*inlh)[q->n].handle, (*inlh)[q->n].assembly_order[j], sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
						ew.w("on \"%s\" (node %d) unable to add sysex input buffer", (*inlh)[q->n].szPname, q->n );
				}
				MIDI_msg m;
				m.node = q->n;
				m.set_sys_ex(tmp,len);
				delete[] tmp;
				if (lp && lp->msg_hnd[MIDI_msg::system])
					lp->msg_hnd[MIDI_msg::system](&m);
				(*inlh)[q->n].assembly_position = 0;
			}
		}
		else
		{
			if (lp && lp->handles(q->dwParam1)) // added this for safety 11/12/95 PP
				lp->parse( q->dwParam1, q->dwParam2, q->n );
		}
		cq.remove_head();
		i++;
	}
	if (i > 0 && statistics_on)
		queueStat.throwInBin(i); // added for statistics gathering 11/12/95 PP
}

void
OMS_app::
write( const MIDI_msg *m )
{
#ifndef NDEBUG
	m->sanityCheck();
#endif
	if (m->ed != NULL) // handle extended message
	{
		if (m->ed->st == smsgt::file_meta)
			return; // ignore meta messages
		schedp->synchronize(); // write_ed is not reentrant
		write_ed(m);
		schedp->unsynchronize();
		return;
	}
	DWORD p;
	UINT n;
	m->to_OMS( &p, &n );
	if (n >= (UINT)(onlh->length))
	{
		ew.w("invalid output node %u",n);
		return;
	}
	if ((*onlh)[n].open == False)
	{
		ew.w("on \"%s\" (node %d) attempt to write output when not open",(*onlh)[n].szPname,n);
		return;
	}
	if ((*onlh)[n].midiator != NULL)
	{
		schedp->synchronize();
		(*onlh)[n].midiator->send(n - (*onlh)[n].midiator->out_start, p);
		schedp->unsynchronize();
		return;
	}
	schedp->synchronize();
	if ((p & 0xf0) == 0xf0 || (*onlh)[n].running_status_time > prev_time) // don't use running status for system messages; plus always throw in status byte every half second
	{
		(*onlh)[n].running_status = 0;
		(*onlh)[n].running_status_time = prev_time + 1000;
	}
	if ((p & 0xff) == (*onlh)[n].running_status)
		p = p >> 8; // get rid of status byte if not needed
	else
		(*onlh)[n].running_status = (uchar)(p & 0xff); // keep track of running status
	MMRESULT mm = midiOutShortMsg((*onlh)[n].handle, p);
	if (mm != MMSYSERR_NOERROR)
	{
		char errtext[MAXERRORLENGTH+1];
		midiOutGetErrorText(mm, errtext, MAXERRORLENGTH);
		ew.w("on \"%s\" (node %d) output error -- %s",(*onlh)[n].szPname,n,errtext);
	}
	schedp->unsynchronize();
}

void
OMS_app::
write_ed(const MIDI_msg *m)
{
/*
	if (m->ed->st == smsgt::file_meta) // just ignore meta messages
	{
//		ew.w("Silly goose, you can't sent a meta message over a MIDI cable!");
		UtilStringPtr s;
		s = m->unparse();
		tw.pr("%s\n",*s);
		s = m->ed->unparse();
		tw.pr("%s\n",*s);
		return;
	}
	else */ if (m->ed->st != smsgt::sys_ex)
	{
		ew.w("attempted to write extended message with unrecognized ed type");
		return;
	}
	UINT n = (UINT)m->node;
	if (n >= (UINT)(onlh->length))
	{
		ew.w("invalid output node %u",n);
		return;
	}
	if ((*onlh)[n].open == False)
	{
		ew.w("on \"%s\" (node %d) attempt to write output when not open",(*onlh)[n].szPname,n);
		return;
	}
	if ((*onlh)[n].midiator != NULL)
	{
		ew.w("on \"%s\" (node %d) system exclusive not supported yet for MIDIator output",(*onlh)[n].szPname,n);
		return;
	}
	(*onlh)[n].running_status = 0; // zap running status so we don't get confused later
	MIDIHDR mh;
	mh.lpData = (LPSTR)(m->ed->sys_ex.d);
	mh.dwBufferLength = m->ed->sys_ex.l;
	mh.dwFlags = 0;
	if (midiOutPrepareHeader((*onlh)[n].handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
	{
		ew.w("on \"%s\" (node %d) unable to prepare MIDIHDR for sysex",(*onlh)[n].szPname,n);
		return;
	}
	if (midiOutLongMsg((*onlh)[n].handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
	{
		ew.w("on \"%s\" (node %d) unable to send sysex",(*onlh)[n].szPname,n);
		midiOutUnprepareHeader((*onlh)[n].handle, &mh, sizeof(MIDIHDR)); // try to clean up
		return;
	}
	DWORD result = WaitForSingleObject(eventHandle, 1000); // wait for sysex to finish
	if (result == WAIT_TIMEOUT)
		ew.w("on \"%s\" (node %d) timed out sending sysex",(*onlh)[n].szPname,n);
	if (midiOutUnprepareHeader((*onlh)[n].handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
	{
		ew.w("on \"%s\" (node %d) unable to unprepare MIDIHDR for sysex",(*onlh)[n].szPname,n);
		return;
	}
}


//--------------------------------------------------------------------------------------------------

#include "string.h" // for strstr()

template <class T>
err_ind
OMS_node_info<T>::
select_by_name( const char *name, TArrayPtr nlh, UINT *id )
{
	UINT num_nodes = nlh->length;

	char name_upper[MAXPNAMELEN+1];
	strtoupper(name, name_upper);

	int none_found = 1;
	
	for (UINT i = 0; i < num_nodes && none_found; i++)
	{
		char tmp[MAXPNAMELEN+1];
		strtoupper( (*nlh)[i].szPname, tmp);
				
		if (strstr(tmp, name_upper))
		{
			tw.pr( "SUCCESS: \"%s\" matched to \"%s\"\n", name, (*nlh)[i].szPname );
			*id = i;
			none_found = 0;
		}
	}
	if (none_found)
		if (num_nodes > 0)
		{	
			ew.w( "Couldn't find \"%s\", using \"%s\" instead", name, (*nlh)[*id].szPname );
		}
		else { ew.w("No input nodes exist"); return err; }
		
	return no_err;
}

template <class T>
void
OMS_node_info<T>::
print_nodes( TArrayPtr nlh, const char *type )
{
	UINT num_nodes = nlh->length;

	tw.pr( "\nnumber of %s nodes: %d\n\n", type, num_nodes );
	for (UINT i=0; i<num_nodes; i++ )
		tw.pr( "%d %s\n", i, (*nlh)[i].szPname );
}

// Authors: Ben Denckla, Herb Yang, Patrick Pelletier
