/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

// This is the main loop for the Rogus Application Framework. It lists all of the
// parsers which have been registered using a ParserRegistration object, and allows
// the user to choose one.

#include "Rogus.h"
#include "OMS_app.h"
#include "text_win.h"
#include "err_and_warn.h"
#include "sched.h"
#include "ParserRegistration.h"
#include "Rogus_McBogus.h"
#include <stdio.h>

//--------------------------------------------------------------------------------------------------

void
main()
{
	char buf[80];
	try
	{
		Rogus_McBogus rogus;
		while (True)
		{
			Sleep(100); // give text window time to catch up
			printf("\nPlease select a parser by choosing a number from below:\n");
			printf("\t 0. exit\n");
			int i;
			int numparsers = ParserRegistration::getNumParsers();
			for (i = 0; i < numparsers; i++)
				printf("\t%2d. %s\n",i+1, ParserRegistration::getParserName(i));
			fgets(buf, 79, stdin);
			int choice = atoi(buf);
			if (choice < 1)
				break;
			choice--;
			if (choice >= numparsers)
			{
				printf("Not a valid choice.\n");
				continue;
			}
			rogus.setParser(ParserRegistration::getParserFunc(choice)(), ParserRegistration::getParserName(choice));
			Sleep(300); // give text window time to catch up
			printf("To choose a new parser, press return.\n");
			fgets(buf, 79, stdin);
			rogus.setParser((parser*)NULL, "");
			oap->closeAllInputs(); // close old parser's inputs
//			oap->closeAllOutputs(); // close old parser's outputs
		}
	}
	catch(err_and_warn::err)
	{
		printf("Rogus McBogus terminated abnormally. Press return to continue\n");
		fgets(buf, 79, stdin);
	}
}

// Authors: Ben Denckla, Patrick Pelletier
