/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef SCORE_H
#define SCORE_H

#include "Filedata.h"
#include "UtilPtr.h"
#include "PerChannel.h"
#include "MIDI_lib.h"
#include "Recorder.h"

class ScoreCreationInfo;
class ScorePlayer;

typedef long ScoreTime;

class Score {
public:
	typedef PerChannel<Bool> ChannelUseMap;
	typedef UtilPtr<ChannelUseMap> ChannelUseMapPtr;
private:
	friend class ScorePlayer;
	typedef UtilArray<MIDI_msg> MIDI_msgArray;
	typedef UtilPtr<MIDI_msgArray> MIDI_msgArrayPtr;
	typedef UtilArray<MIDI_msgArrayPtr> MIDI_msgArrayPtrArray;
	typedef UtilPtr<MIDI_msgArrayPtrArray> MIDI_msgArrayPtrArrayPtr;
	typedef void (*ScoreFunc)(void*, ScorePlayer*);
	typedef UtilArray<ScoreFunc> ScoreFuncArray;
	typedef UtilPtr<ScoreFuncArray> ScoreFuncArrayPtr;
	typedef void* VoidStar;
	typedef UtilArray<VoidStar> VoidStarArray;
	typedef UtilPtr<VoidStarArray> VoidStarArrayPtr;
	typedef UtilArray<ScoreTime> ScoreTimeArray;
	typedef UtilPtr<ScoreTimeArray> ScoreTimeArrayPtr;
	MIDI_msgArrayPtr msgs;
	ChannelUseMapPtr u;
	ScoreFuncArrayPtr funcs;
	VoidStarArrayPtr args;
	ScoreTimeArrayPtr times;
	void read(const Filedata*);
	static void warnfunc(const char *);
	static void convertTime(MIDI_msgArrayPtr, long, ScoreTimeArrayPtr);
	static void convertFuncTime(ScoreTimeArrayPtr stap, ScoreTime lastTime, long lastTick, long thisTick, int* func_index, long currentTempo, long division);
public:
	Score() { msgs = new MIDI_msgArray(0); }
	Score(char* s) { Filedata f(s, warnfunc); read(&f); }
	Score(const Filedata* f) { read(f); }
	Score(const ScoreCreationInfo*);
	Score(const Recorder*);
	int getNumEvents() const { return msgs->length; }
	ScoreTime getLength() const;
	ChannelUseMapPtr getChannels() const;
	const MIDI_msg& operator[](int i) const { return (*(((Score*)this)->msgs))[i]; }
};

#endif
// Author: Patrick Pelletier
