/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

 /* This file gives example of how to create a parser with hooks into your own MIDI input 
processing functions.  In this case, a delayed octave generator is made. */

#include "octave_parser.h"
#include "data_structs.h"
#include "ParserRegistration.h"

static int ou_node;
static pool<MIDI_msg> b(50, "");
static Bool stop[128];

static const int note_len = 250, delay = 150;

static long wake( void *arg )
{
	MIDI_msg* m = (MIDI_msg*) arg;
	if (stop[m->key()])
	{
		b.release(m);
		return NEVER;
	}
	else
	{
		schedp->play_note( *m, note_len );
		return delay;
	}
}

static
void
note_on( MIDI_msg *m ) 
{
	m->node = ou_node;
	if (m->vel())
	{
		schedp->play_note( *m, note_len );
		
		stop[m->key()] = 0;

		MIDI_msg *n = b.allocate();
		if (!n)
			ew.w( "allocate failed" );
		else
		{
			*n = *m;

			task_data td( wake, n, m->time + delay );
			int e = schedp->insert( td );
			if (e) ew.w( "insert failed in octave parser" );
		}
	}
	else stop[m->key()] = 1;
}

octave_parser::
octave_parser()
{
	ou_node = oap->open_output(); // open the default output, and save its ioref
	oap->open_input(); // open the default input

	msg_hnd[MIDI_msg::note_on] = note_on;
		
	MIDI_msg m;
	m.set( MIDI_msg::program_chn, 0 /* chan */, 20, 0 /* not used */, ou_node );
	//oap->write(&m);
}

static parser* newFunc()
{
	return (new octave_parser());
}

static ParserRegistration reg(newFunc, "octave_parser");

// Author: Ben Denckla
